/*
 *	CC Network Filesystem (ccgfs)
 *	Storage Engine
 *
 *	Copyright © Jan Engelhardt <jengelh [at] medozas de>, 2007-2011
 *
 *	This file is part of CCGFS. CCGFS is free software; you can
 *	redistribute it and/or modify it under the terms of the GNU
 *	General Public License as published by the Free Software
 *	Foundation; either version 2 or 3 of the License.
 */
#define _ATFILE_SOURCE 1
#define _GNU_SOURCE 1
#include "config.h"
#ifdef HAVE_SYS_FSUID_H
#	include <sys/fsuid.h>
#endif
#ifdef HAVE_SYS_MMAN_H
#	include <sys/mman.h>
#endif
#include <sys/stat.h>
#ifdef HAVE_SYS_STATVFS_H
#	include <sys/statvfs.h>
#endif
#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#ifdef HAVE_XATTR_XATTR_H
#	include <attr/xattr.h>
#endif
#ifdef HAVE_SYS_XATTR_H
#	include <sys/xattr.h>
#endif
#include <libHX/init.h>
#include <libHX/option.h>
#include "ccgfs.h"
#include "config.h"
#include "packet.h"
#include "xl.h"
#ifdef ENABLE_WORKAROUNDS
#	include "xat.h"
#endif

#define b_path(dest, src) /* build path */ \
	(snprintf(dest, sizeof(dest), "%s%s", root_dir, (src)) >= \
	          sizeof(dest))
#define perror(s) perror("ccgfs-storage: " s)

enum {
	LOCALFS_SUCCESS = 0,
	LOCALFS_STOP,
};

typedef int (*localfs_func_t)(int, struct lo_packet *);

static char root_dir[PATH_MAX];
static int root_fd;
static unsigned int pagesize, exact_uid_mapping = true;

static __attribute__((pure)) const char *at(const char *in)
{
	if (*in != '/')
		abort();
	if (in[1] == '\0')
		return ".";
	return in + 1;
}

static int localfs_chmod(int fd, struct lo_packet *rq)
{
	const char *rq_path = pkt_shift_s(rq);
	mode_t rq_mode      = pkt_shift_32(rq);

	if (fchmodat(root_fd, at(rq_path), rq_mode, 0) < 0)
		return -errno;

	return LOCALFS_SUCCESS;
}

static int localfs_chown(int fd, struct lo_packet *rq)
{
	const char *rq_path = pkt_shift_s(rq);
	uid_t rq_uid        = pkt_shift_32(rq);
	gid_t rq_gid        = pkt_shift_32(rq);

	if (fchownat(root_fd, at(rq_path), rq_uid, rq_gid,
	    AT_SYMLINK_NOFOLLOW) < 0)
		return -errno;

	return LOCALFS_SUCCESS;
}

static int localfs_create(int fd, struct lo_packet *rq)
{
	const char *rq_path   = pkt_shift_s(rq);
	unsigned int rq_flags = pkt_shift_32(rq);
	mode_t rq_mode        = pkt_shift_32(rq);
	struct lo_packet *rp;
	int ret;

	ret = openat(root_fd, at(rq_path), arch_openflags(rq_flags), rq_mode);
	if (ret < 0)
		return -errno;

	rp = pkt_init(CCGFS_CREATE_RESPONSE, PV_32);
	pkt_push_32(rp, ret);
	pkt_send(fd, rp);
	return LOCALFS_STOP;
}

static struct lo_packet *getattr_copy_stor(const struct stat *sb, uint8_t type)
{
	struct lo_packet *rp;

	rp = pkt_init(type, 7 * PV_64 + 8 * PV_32);
	if (rp == NULL)
		return NULL;

	pkt_push_64(rp, sb->st_ino);
	pkt_push_32(rp, sb->st_mode);
	pkt_push_32(rp, sb->st_nlink);
	pkt_push_32(rp, sb->st_uid);
	pkt_push_32(rp, sb->st_gid);
	pkt_push_32(rp, sb->st_rdev);
	pkt_push_64(rp, sb->st_size);
	pkt_push_64(rp, sb->st_blksize);
	pkt_push_64(rp, sb->st_blocks);
	if (type == CCGFS_GETATTR2_RESPONSE) {
#if defined(HAVE_STRUCT_STAT_ST_MTIM)
		pkt_push_64(rp, sb->st_atim.tv_sec);
		pkt_push_32(rp, sb->st_atim.tv_nsec);
		pkt_push_64(rp, sb->st_mtim.tv_sec);
		pkt_push_32(rp, sb->st_mtim.tv_nsec);
		pkt_push_64(rp, sb->st_ctim.tv_sec);
		pkt_push_32(rp, sb->st_ctim.tv_nsec);
#elif defined(HAVE_STRUCT_STAT_ST_MTIMENSEC)
		pkt_push_64(rp, sb->st_atime);
		pkt_push_32(rp, sb->st_atimensec);
		pkt_push_64(rp, sb->st_mtime);
		pkt_push_32(rp, sb->st_mtimensec);
		pkt_push_64(rp, sb->st_ctime);
		pkt_push_32(rp, sb->st_ctimensec);
#else
		pkt_push_64(rp, sb->st_atime);
		pkt_push_32(rp, 0);
		pkt_push_64(rp, sb->st_mtime);
		pkt_push_32(rp, 0);
		pkt_push_64(rp, sb->st_ctime);
		pkt_push_32(rp, 0);
#endif
	} else if (type == CCGFS_GETATTR_RESPONSE) {
		pkt_push_64(rp, sb->st_atime);
		pkt_push_64(rp, sb->st_mtime);
		pkt_push_64(rp, sb->st_ctime);
	}
	return rp;
}

static int localfs_fgetattr(int fd, struct lo_packet *rq)
{
	int rq_fd = pkt_shift_32(rq);
	struct stat sb;

	if (fstat(rq_fd, &sb) < 0)
		return -errno;

	pkt_send(fd, getattr_copy_stor(&sb, CCGFS_GETATTR_RESPONSE));
	return LOCALFS_STOP;
}

static int localfs_fgetattr2(int fd, struct lo_packet *rq)
{
	int rq_fd = pkt_shift_32(rq);
	struct stat sb;

	if (fstat(rq_fd, &sb) < 0)
		return -errno;

	pkt_send(fd, getattr_copy_stor(&sb, CCGFS_GETATTR2_RESPONSE));
	return LOCALFS_STOP;
}

static int localfs_fsync(int fd, struct lo_packet *rq)
{
	int rq_fd              = pkt_shift_32(rq);
	unsigned int data_only = pkt_shift_32(rq);

	if (data_only) {
		if (fdatasync(rq_fd) < 0)
			return -errno;
	} else {
		if (fsync(rq_fd) < 0)
			return -errno;
	}

	return LOCALFS_SUCCESS;
}

static int localfs_ftruncate(int fd, struct lo_packet *rq)
{
	int rq_fd    = pkt_shift_32(rq);
	off_t rq_off = pkt_shift_64(rq);

	if (ftruncate(rq_fd, rq_off) < 0)
		return -errno;

	return LOCALFS_SUCCESS;
}

static int localfs_getattr(int fd, struct lo_packet *rq)
{
	const char *rq_path = pkt_shift_s(rq);
	struct stat sb;

	if (fstatat(root_fd, at(rq_path), &sb, AT_SYMLINK_NOFOLLOW) < 0)
		return -errno;

	pkt_send(fd, getattr_copy_stor(&sb, CCGFS_GETATTR_RESPONSE));
	return LOCALFS_STOP;
}

static int localfs_getattr2(int fd, struct lo_packet *rq)
{
	const char *rq_path = pkt_shift_s(rq);
	struct stat sb;

	if (fstatat(root_fd, at(rq_path), &sb, AT_SYMLINK_NOFOLLOW) < 0)
		return -errno;

	pkt_send(fd, getattr_copy_stor(&sb, CCGFS_GETATTR2_RESPONSE));
	return LOCALFS_STOP;
}

static int localfs_getxattr(int fd, struct lo_packet *rq)
{
	const char *rq_path = pkt_shift_s(rq);
	const char *rq_name = pkt_shift_s(rq);
	size_t rq_size      = pkt_shift_64(rq);
	struct lo_packet *rp;
	ssize_t ret;
	void *value;

	ret = lgetxattr(at(rq_path), rq_name, NULL, 0);
	if (ret < 0)
		return -errno;
	if (rq_size == 0) {
		rp = pkt_init(CCGFS_GETXATTR_RESPONSE, PV_64 + PV_DATA(0));
		pkt_push_64(rp, ret);
		pkt_push(rp, NULL, 0, PT_DATA);
		pkt_send(fd, rp);
		return LOCALFS_STOP;
	}
	value = malloc(rq_size);
	if (value == NULL)
		return -ENOMEM;
	ret = lgetxattr(at(rq_path), rq_name, value, rq_size);
	if (ret < 0) {
		ret = -errno;
		free(value);
		return ret;
	}
	rp = pkt_init(CCGFS_GETXATTR_RESPONSE, PT_64 + PV_DATA(ret));
	pkt_push_64(rp, ret);
	pkt_push(rp, value, ret, PT_DATA);
	pkt_send(fd, rp);
	free(value);
	return LOCALFS_STOP;
}

static int localfs_link(int fd, struct lo_packet *rq)
{
	const char *rq_oldpath = pkt_shift_s(rq);
	const char *rq_newpath = pkt_shift_s(rq);

	if (linkat(root_fd, at(rq_oldpath), root_fd, at(rq_newpath), 0) < 0)
		return -errno;

	return LOCALFS_SUCCESS;
}

static int localfs_listxattr(int fd, struct lo_packet *rq)
{
	const char *rq_path = pkt_shift_s(rq);
	size_t rq_size      = pkt_shift_64(rq);
	struct lo_packet *rp;
	ssize_t ret;
	char *list;

	ret = llistxattr(at(rq_path), NULL, 0);
	if (ret < 0)
		return -errno;
	if (rq_size == 0) {
		rp = pkt_init(CCGFS_LISTXATTR_RESPONSE, PV_64 + PV_DATA(0));
		pkt_push_64(rp, ret);
		pkt_push(rp, NULL, 0, PT_DATA);
		pkt_send(fd, rp);
		return LOCALFS_STOP;
	}
	list = malloc(ret);
	if (list == NULL)
		return -ENOMEM;
	ret = llistxattr(at(rq_path), list, rq_size);
	if (ret < 0) {
		ret = -errno;
		free(list);
		return ret;
	}
	rp = pkt_init(CCGFS_LISTXATTR_RESPONSE, PT_64 + PV_DATA(ret));
	pkt_push_64(rp, ret);
	pkt_push(rp, list, ret, PT_DATA);
	pkt_send(fd, rp);
	free(list);
	return LOCALFS_STOP;
}

static int localfs_mkdir(int fd, struct lo_packet *rq)
{
	const char *rq_path = pkt_shift_s(rq);
	mode_t rq_mode      = pkt_shift_32(rq);

	if (mkdirat(root_fd, at(rq_path), rq_mode) < 0)
		return -errno;

	return LOCALFS_SUCCESS;
}

static int localfs_mknod(int fd, struct lo_packet *rq)
{
	const char *rq_path = pkt_shift_s(rq);
	mode_t rq_mode      = pkt_shift_32(rq);
	dev_t rq_rdev       = pkt_shift_32(rq);

	if (mknodat(root_fd, at(rq_path), rq_mode, rq_rdev) < 0)
		return -errno;

	return LOCALFS_SUCCESS;
}

static int localfs_open(int fd, struct lo_packet *rq)
{
	const char *rq_path   = pkt_shift_s(rq);
	unsigned int rq_flags = pkt_shift_32(rq);
	struct lo_packet *rp;
	int ret;

	ret = openat(root_fd, at(rq_path), arch_openflags(rq_flags));
	if (ret < 0)
		return -errno;

	rp = pkt_init(CCGFS_OPEN_RESPONSE, PV_32);
	pkt_push_32(rp, ret);
	pkt_send(fd, rp);
	return LOCALFS_STOP;
}

static int localfs_opendir_access(int fd, struct lo_packet *rq)
{
	const char *rq_path = pkt_shift_s(rq);
	struct stat sb;

	if (fstatat(root_fd, at(rq_path), &sb, AT_SYMLINK_NOFOLLOW) < 0)
		return -errno;
	if (!S_ISDIR(sb.st_mode))
		return -ENOTDIR;
	//setreuid(fsuid, -1);
	if (faccessat(root_fd, at(rq_path), X_OK, AT_SYMLINK_NOFOLLOW) < 0)
		return -errno;
	return LOCALFS_SUCCESS;
}

static int localfs_read(int fd, struct lo_packet *rq)
{
	int rq_fd       = pkt_shift_32(rq);
	size_t rq_size  = pkt_shift_64(rq);
	off_t rq_offset = pkt_shift_64(rq);

	struct lo_packet *rp;
	ssize_t ret;
	char *buf;

	buf = malloc(rq_size);
	if (buf == NULL)
		return -EIO;
	ret = pread(rq_fd, buf, rq_size, rq_offset);
	if (ret < 0) {
		if (errno != ESPIPE) {
			free(buf);
			return -errno;
		}
		ret = read(rq_fd, buf, rq_size);
	}
	if (ret < 0) {
		free(buf);
		return -errno;
	}

	rp = pkt_init(CCGFS_READ_RESPONSE, PV_64 + PV_DATA(ret));
	pkt_push_64(rp, ret);
	pkt_push(rp, buf, ret, PT_DATA);
	pkt_send(fd, rp);
	free(buf);
	return LOCALFS_STOP;
}

static int localfs_readdir(int fd, struct lo_packet *rq)
{
	const char *rq_path = pkt_shift_s(rq);
	struct dirent *dentry;
	struct lo_packet *rp;
	DIR *ptr;

	if ((ptr = opendir(at(rq_path))) == NULL)
		return -errno;

	while ((dentry = readdir(ptr)) != NULL) {
		rp = pkt_init(CCGFS_READDIR_RESPONSE,
		              PV_64 + PV_32 + PV_STRZ(dentry->d_name));
		pkt_push_64(rp, dentry->d_ino);
		pkt_push_32(rp, dentry->d_type);
		pkt_push_s(rp, dentry->d_name);
		pkt_send(fd, rp);
	}

	closedir(ptr);
	return LOCALFS_SUCCESS;
}

static int localfs_readlink(int fd, struct lo_packet *rq)
{
	const char *rq_path = pkt_shift_s(rq);
	char d_linkbuf[PATH_MAX];
	struct lo_packet *rp;

	memset(d_linkbuf, 0, sizeof(d_linkbuf));
	if (readlinkat(root_fd, at(rq_path), d_linkbuf,
	    sizeof(d_linkbuf) - 1) < 0)
		return -errno;

	rp = pkt_init(CCGFS_READLINK_RESPONSE, PV_STRZ(d_linkbuf));
	pkt_push_s(rp, d_linkbuf);
	pkt_send(fd, rp);
	return LOCALFS_STOP;
}

static int localfs_release(int fd, struct lo_packet *rq)
{
	if (close(pkt_shift_32(rq)) < 0)
		return -errno;
	return LOCALFS_SUCCESS;
}

static int localfs_removexattr(int fd, struct lo_packet *rq)
{
	const char *rq_path = pkt_shift_s(rq);
	const char *rq_name = pkt_shift_s(rq);

	if (lremovexattr(at(rq_path), rq_name) < 0)
		return -errno;

	return LOCALFS_SUCCESS;
}

static int localfs_rename(int fd, struct lo_packet *rq)
{
	const char *rq_oldpath = pkt_shift_s(rq);
	const char *rq_newpath = pkt_shift_s(rq);

	if (renameat(root_fd, at(rq_oldpath), root_fd, at(rq_newpath)) < 0)
		return -errno;

	return LOCALFS_SUCCESS;
}

static int localfs_rmdir(int fd, struct lo_packet *rq)
{
	const char *rq_path = pkt_shift_s(rq);

	if (unlinkat(root_fd, at(rq_path), AT_REMOVEDIR) < 0)
		return -errno;

	return LOCALFS_SUCCESS;
}

static int localfs_setxattr(int fd, struct lo_packet *rq)
{
	const char *rq_path  = pkt_shift_s(rq);
	const char *rq_name  = pkt_shift_s(rq);
	const void *rq_value = pkt_shift_s(rq);
	size_t rq_size       = pkt_shift_64(rq);
	unsigned int flags   = pkt_shift_32(rq);

	if (lsetxattr(at(rq_path), rq_name, rq_value, rq_size, flags) < 0)
		return -errno;

	return LOCALFS_SUCCESS;
}

static int localfs_symlink(int fd, struct lo_packet *rq)
{
	const char *rq_oldpath = pkt_shift_s(rq);
	const char *rq_newpath = pkt_shift_s(rq);

	if (symlinkat(rq_oldpath, root_fd, at(rq_newpath)) < 0)
		return -errno;

	return LOCALFS_SUCCESS;
}

static int localfs_statfs(int fd, struct lo_packet *rq)
{
	struct lo_packet *rp;
	struct statvfs st;

	if (fstatvfs(root_fd, &st) < 0)
		return -errno;

	rp = pkt_init(CCGFS_STATFS_RESPONSE, 11 * PV_64);
	pkt_push_64(rp, st.f_bsize);
	pkt_push_64(rp, st.f_frsize);
	pkt_push_64(rp, st.f_blocks);
	pkt_push_64(rp, st.f_bfree);
	pkt_push_64(rp, st.f_bavail);
	pkt_push_64(rp, st.f_files);
	pkt_push_64(rp, st.f_ffree);
	pkt_push_64(rp, st.f_favail);
	pkt_push_64(rp, st.f_fsid);
	pkt_push_64(rp, st.f_flag);
	pkt_push_64(rp, st.f_namemax);
	pkt_send(fd, rp);
	return LOCALFS_STOP;
}

static int localfs_truncate(int fd, struct lo_packet *rq)
{
	const char *rq_path = pkt_shift_s(rq);
	off_t rq_off        = pkt_shift_64(rq);

	if (truncate(at(rq_path), rq_off) < 0)
		return -errno;

	return LOCALFS_SUCCESS;
}

static int localfs_unlink(int fd, struct lo_packet *rq)
{
	const char *rq_path = pkt_shift_s(rq);

	if (unlinkat(root_fd, at(rq_path), 0) < 0)
		return -errno;

	return LOCALFS_SUCCESS;
}

static int localfs_utimens(int fd, struct lo_packet *rq)
{
	const char *rq_path = pkt_shift_s(rq);
	struct timespec val[2];

	val[0].tv_sec  = pkt_shift_64(rq);
	val[0].tv_nsec = pkt_shift_64(rq);
	val[1].tv_sec  = pkt_shift_64(rq);
	val[1].tv_nsec = pkt_shift_64(rq);
	if (utimensat(root_fd, at(rq_path), val, AT_SYMLINK_NOFOLLOW) < 0)
		return -errno;

	return LOCALFS_SUCCESS;
}

static int localfs_write(int fd, struct lo_packet *rq)
{
	int rq_fd        = pkt_shift_32(rq);
	size_t size      = pkt_shift_64(rq);
	off_t offset     = pkt_shift_64(rq);
	const void *data = pkt_shift_s(rq);

	struct lo_packet *rp;
	ssize_t ret;

	ret = pwrite(rq_fd, data, size, offset);
	if (ret < 0) {
		if (errno != ESPIPE)
			return -errno;
		ret = write(rq_fd, data, size);
	}
	if (ret < 0)
		return -errno;

	rp = pkt_init(CCGFS_ERRNO_RESPONSE, PV_32);
	pkt_push_32(rp, ret);
	pkt_send(fd, rp);
	return LOCALFS_STOP;
}

static const localfs_func_t localfs_func_array[] = {
	[CCGFS_CHMOD_REQUEST]       = localfs_chmod,
	[CCGFS_CHOWN_REQUEST]       = localfs_chown,
	[CCGFS_CREATE_REQUEST]      = localfs_create,
	[CCGFS_FGETATTR_REQUEST]    = localfs_fgetattr,
	[CCGFS_FGETATTR2_REQUEST]   = localfs_fgetattr2,
	[CCGFS_FSYNC_REQUEST]       = localfs_fsync,
	[CCGFS_FTRUNCATE_REQUEST]   = localfs_ftruncate,
	[CCGFS_GETATTR_REQUEST]     = localfs_getattr,
	[CCGFS_GETATTR2_REQUEST]    = localfs_getattr2,
	[CCGFS_GETXATTR_REQUEST]    = localfs_getxattr,
	[CCGFS_LINK_REQUEST]        = localfs_link,
	[CCGFS_LISTXATTR_REQUEST]   = localfs_listxattr,
	[CCGFS_MKDIR_REQUEST]       = localfs_mkdir,
	[CCGFS_MKNOD_REQUEST]       = localfs_mknod,
	[CCGFS_OPEN_REQUEST]        = localfs_open,
	[CCGFS_OPENDIR_REQUEST]     = localfs_opendir_access,
	[CCGFS_READ_REQUEST]        = localfs_read,
	[CCGFS_READDIR_REQUEST]     = localfs_readdir,
	[CCGFS_READLINK_REQUEST]    = localfs_readlink,
	[CCGFS_RELEASE_REQUEST]     = localfs_release,
	[CCGFS_REMOVEXATTR_REQUEST] = localfs_removexattr,
	[CCGFS_RENAME_REQUEST]      = localfs_rename,
	[CCGFS_RMDIR_REQUEST]       = localfs_rmdir,
	[CCGFS_SETXATTR_REQUEST]    = localfs_setxattr,
	[CCGFS_STATFS_REQUEST]      = localfs_statfs,
	[CCGFS_SYMLINK_REQUEST]     = localfs_symlink,
	[CCGFS_TRUNCATE_REQUEST]    = localfs_truncate,
	[CCGFS_UNLINK_REQUEST]      = localfs_unlink,
	[CCGFS_UTIMENS_REQUEST]     = localfs_utimens,
	[CCGFS_WRITE_REQUEST]       = localfs_write,
};

static int localfs_setfsid(struct lo_packet *rq)
{
	uid_t uid = pkt_shift_32(rq);
	gid_t gid = pkt_shift_32(rq);
	if (!exact_uid_mapping)
		return 0;
	/*
	 * setfsuid has no proper error codes, aagh.. But at least we warned
	 * in get_options when setfsuid won't work.
	 */
	setfsuid(uid);
	setfsgid(gid);
	return 0;
}

static void handle_packet(int fd, struct lo_packet *rq)
{
	struct ccgfs_pkt_header *hdr;
	struct lo_packet *rp;
	localfs_func_t lf;
	int ret;

	ret = localfs_setfsid(rq);
	if (ret < 0) {
		rp = pkt_init(CCGFS_ERRNO_RESPONSE, PV_32);
		pkt_push_32(rp, generic_errno(ret));
		pkt_send(fd, rp);
		return;
	}

	ret = -EIO;
	hdr = rq->data;
	lf  = localfs_func_array[hdr->opcode];
	if (lf != NULL)
		ret = (*lf)(fd, rq);

	if (ret <= 0) {
		rp = pkt_init(CCGFS_ERRNO_RESPONSE, PV_32);
		pkt_push_32(rp, generic_errno(ret));
		pkt_send(fd, rp);
	}
}

static void send_fsinfo(int fd)
{
	struct lo_packet *rp;
	char host[NAME_MAX], buf[NAME_MAX];

	if (gethostname(host, sizeof(host)) < 0) {
		perror("gethostname");
		exit(EXIT_FAILURE);
	}

	snprintf(buf, sizeof(buf), "%s:%s", host, root_dir);
	rp = pkt_init(CCGFS_FSINFO, PV_STRZ(buf));
	pkt_push_s(rp, buf);
	pkt_send(fd, rp);
}

static bool get_options(int *argc, const char ***argv)
{
	static const struct HXoption options_table[] = {
		{.sh = 'P', .type = HXTYPE_VAL, .ptr = &exact_uid_mapping,
		 .val = true, .help = "Enable 1:1 UID mapping"},
		{.sh = 'p', .type = HXTYPE_VAL, .ptr = &exact_uid_mapping,
		 .val = false, .help = "Disable 1:1 UID mapping"},
		HXOPT_AUTOHELP,
		HXOPT_TABLEEND,
	};
	if (HX_getopt(options_table, argc, argv, HXOPT_USAGEONERR) !=
	    HXOPT_ERR_SUCCESS)
		return false;

	if (exact_uid_mapping) {
		/* Odd setfsuid semantics require calling it over again. */
		setfsuid(geteuid() + 1);
		if (setfsuid(geteuid() + 1) == geteuid())
			fprintf(stderr, "%s: warning: you do not seem to have "
			        "permission to use setfsuid. This renders "
			        "1:1 UID mapping ineffective.\n", **argv);
		setfsuid(geteuid());
	}

	return true;
}

static int main2(int argc, const char **argv)
{
	struct lo_packet *rq;

	if (!get_options(&argc, &argv))
		return EXIT_FAILURE;
	if (argc < 2) {
		fprintf(stderr, "%s: directory to be exported is "
		        "needed\n", *argv);
		return EXIT_FAILURE;
	}

	if ((root_fd = open(argv[1], O_DIRECTORY)) < 0) {
		fprintf(stderr, "ccgfs-storage: could not open \"%s\": %s\n",
		        argv[1], strerror(errno));
		return EXIT_FAILURE;
	}
	if (fchdir(root_fd) < 0) {
		perror("chdir");
		return EXIT_FAILURE;
	}
	if (getcwd(root_dir, sizeof(root_dir)) == NULL) {
		perror("getcwd");
		return EXIT_FAILURE;
	}

	umask(0);
	pagesize  = sysconf(_SC_PAGESIZE);
	send_fsinfo(STDOUT_FILENO);

	while (true) {
		rq = pkt_recv(STDIN_FILENO);
		if (rq == NULL)
			break;
		handle_packet(STDOUT_FILENO, rq);
		pkt_destroy(rq);
	}

	return EXIT_SUCCESS;
}

int main(int argc, const char **argv)
{
	int ret;

	ret = HX_init();
	if (ret <= 0) {
		fprintf(stderr, "HX_init: %s\n", strerror(errno));
		abort();
	}

	pktlayer_init();
	ret = main2(argc, argv);
	HX_exit();
	return ret;
}
